<?php
namespace Craft;

/**
 *
 */
class UserSessionService extends \CWebUser
{
	const FLASH_KEY_PREFIX = 'Craft.UserSessionService.flash.';
	const FLASH_COUNTERS   = 'Craft.UserSessionService.flashcounters';

	/**
	 * Stores the user identity.
	 *
	 * @access private
	 * @var UserIdentity
	 */
	private $_identity;

	/**
	 * Stores the current user model.
	 *
	 * @access private
	 * @var UserModel
	 */
	private $_userModel;

	/**
	 * @var
	 */
	private $_userRow;

	/**
	 * @var
	 */
	private $_sessionRestoredFromCookie;

	/**
	 *
	 */
	function __construct()
	{
		// If the identity cookie is missing we assume it has expired.  We need to kill the PHP session information as early
		// as possible in the request if the session duration is set to be anything greater than 0.
		$cookies = craft()->request->getCookies();
		$cookie = $cookies->itemAt($this->getStateKeyPrefix());

		// If there is no cookie, then assume it has expired
		if (!$cookie)
		{
			// If session duration is set to 0, then the session will be over when the browser is closed.
			if ($this->_getSessionDuration(false) > 0)
			{
				// No soup for you!
				$this->logout(false);
			}
		}
	}

	/**
	 * Gets the currently logged-in user.
	 *
	 * @return UserModel|null
	 */
	public function getUser()
	{
		// Does a user appear to be logged in?
		if (Craft::isInstalled() && $this->getState('__id') !== null)
		{
			if (!isset($this->_user))
			{
				$userRow = $this->_getUserRow($this->getId());

				if ($userRow)
				{
					$this->_userModel = UserModel::populateModel($userRow);
				}
				else
				{
					$this->_userModel = false;
				}
			}

			return $this->_userModel ? $this->_userModel : null;
		}
	}

	/**
	 * Returns the URL the user was trying to access before getting sent to the login page.
	 *
	 * @param string $defaultUrl
	 * @return mixed
	 */
	public function getReturnUrl($defaultUrl = '')
	{
		return $this->getState('__returnUrl', UrlHelper::getUrl($defaultUrl));
	}

	/**
	 * Sets a notice to the user.
	 *
	 * @param string $message
	 */
	public function setNotice($message)
	{
		$this->setFlash('notice', $message);
	}

	/**
	 * Sets an error notification.
	 *
	 * @param string $message
	 */
	public function setError($message)
	{
		$this->setFlash('error', $message);
	}

	/**
	 *
	 * Check to see if the current web user is a guest.
	 *
	 * (wrapper for getIsGuest() for consistency)
	 *
	 * @return bool
	 */
	public function isGuest()
	{
		$user = $this->_getUserRow($this->getId());
		return empty($user);
	}

	/**
	 * Check to see if the current web user is logged in.
	 *
	 * @return bool
	 */
	public function isLoggedIn()
	{
		return !$this->isGuest();
	}

	/**
	 * Returns whether the current user is an admin.
	 *
	 * @return bool
	 */
	public function isAdmin()
	{
		$user = $this->getUser();
		return ($user && $user->admin);
	}

	/**
	 * Returns whether the current user has a given permission.
	 *
	 * @param string $permissionName
	 * @return bool
	 */
	public function checkPermission($permissionName)
	{
		$user = $this->getUser();
		return ($user && $user->can($permissionName));
	}

	/**
	 * Requires that the current user has a given permission, otherwise a 403 exception is thrown.
	 *
	 * @param string $permissionName
	 * @throws HttpException
	 */
	public function requirePermission($permissionName)
	{
		if (!$this->checkPermission($permissionName))
		{
			throw new HttpException(403);
		}
	}

	/**
	 * Requires that the current user is an admin, otherwise a 403 exception is thrown.
	 *
	 * @throws HttpException
	 */
	public function requireAdmin()
	{
		if (!$this->isAdmin())
		{
			throw new HttpException(403);
		}
	}

	/**
	 * Requires that the user is logged in, otherwise redirects them to the login page.
	 */
	public function requireLogin()
	{
		if ($this->isGuest())
		{
			if (!craft()->request->isAjaxRequest())
			{
				if (craft()->request->getPathInfo() !== '')
				{
					$this->setReturnUrl(craft()->request->getPath());
				}
			}
			elseif (isset($this->loginRequiredAjaxResponse))
			{
				echo $this->loginRequiredAjaxResponse;
				craft()->end();
			}

			$url = UrlHelper::getUrl($this->loginUrl);
			craft()->request->redirect($url);
		}
	}

	/**
	 * Pointless Wrapper for requireLogin(), but \CWebUser uses loginRequired() so we must support it as well.
	 */
	public function loginRequired()
	{
		$this->requireLogin();
	}

	/**
	 * Logs a user in.
	 *
	 * @param \IUserIdentity $username
	 * @param int            $password
	 * @param bool           $rememberMe
	 * @throws Exception
	 * @return bool
	 */
	public function login($username, $password, $rememberMe = false)
	{
		// Validate the username/password first.
		$usernameModel = new UsernameModel();
		$passwordModel = new PasswordModel();

		$usernameModel->username = $username;
		$passwordModel->password = $password;

		// Require a userAgent string and an IP address to help prevent direct socket connections from trying to login.
		if (!craft()->request->userAgent || !craft()->request->getIpAddress())
		{
			Craft::log('Someone tried to login with loginName: '.$username.', without presenting an IP address or userAgent string.', LogLevel::Warning);
			$this->logout();
			$this->requireLogin();
		}

		// Validate the model.
		if ($usernameModel->validate() && $passwordModel->validate())
		{
			// Authenticate the credentials.
			$this->_identity = new UserIdentity($username, $password);
			$this->_identity->authenticate();

			// Was the login successful?
			if ($this->_identity->errorCode == UserIdentity::ERROR_NONE)
			{
				// See if the 'rememberUsernameDuration' config item is set. If so, save the name to a cookie.
				$rememberUsernameDuration = craft()->config->get('rememberUsernameDuration');
				if ($rememberUsernameDuration)
				{
					$interval = new DateInterval($rememberUsernameDuration);
					$expire = new DateTime();
					$expire->add($interval);

					// Save the username cookie.
					$this->saveCookie('username', $username, $expire->getTimestamp());
				}

				// Get how long this session is supposed to last.
				$seconds = $this->_getSessionDuration($rememberMe);
				$this->authTimeout = $seconds;

				$id = $this->_identity->getId();
				$states = $this->_identity->getPersistentStates();

				// Run any before login logic.
				if ($this->beforeLogin($id, $states, false))
				{
					$this->changeIdentity($id, $this->_identity->getName(), $states);

					if ($seconds > 0)
					{
						if ($this->allowAutoLogin)
						{
							$user = craft()->users->getUserById($id);

							if ($user)
							{
								// Save the necessary info to the identity cookie.
								$sessionToken = StringHelper::UUID();
								$hashedToken = craft()->security->hashString($sessionToken);
								$uid = craft()->users->handleSuccessfulLogin($user, $hashedToken['hash']);
								$userAgent = craft()->request->userAgent;

								$data = array(
									$this->getName(),
									$sessionToken,
									$uid,
									$seconds,
									$userAgent,
									$this->saveIdentityStates(),
								);

								$this->saveCookie('', $data, $seconds);
							}
							else
							{
								throw new Exception(Craft::t('Could not find a user with Id of {userId}.', array('{userId}' => $this->getId())));
							}
						}
						else
						{
							throw new Exception(Craft::t('{class}.allowAutoLogin must be set true in order to use cookie-based authentication.', array('{class}' => get_class($this))));
						}
					}

					$this->_sessionRestoredFromCookie = false;
					$this->_userRow = null;

					// Run any after login logic.
					$this->afterLogin(false);
				}

				return !$this->getIsGuest();
			}
		}

		Craft::log($username.' tried to log in unsuccessfully.', LogLevel::Warning);
		return false;
	}

	/**
	 * Returns the login error code from the user identity.
	 *
	 * @return UserIdentity
	 */
	public function getLoginErrorCode()
	{
		if (isset($this->_identity))
		{
			return $this->_identity->errorCode;
		}
	}

	/**
	 * @return string
	 */
	public function getRememberedUsername()
	{
		return $this->getCookieValue('username');
	}

	/**
	 * Overriding Yii's because it's stupid.
	 *
	 * @return bool
	 */
	public function getIsGuest()
	{
		return $this->isGuest();
	}

	/**
	 * TODO: set domain to wildcard?  .example.com, .example.co.uk, .too.many.subdomains.com
	 *
	 * @param     $cookieName
	 * @param     $data
	 * @param int $duration
	 */
	public function saveCookie($cookieName, $data, $duration = 0)
	{
		$cookieName = $this->getStateKeyPrefix().$cookieName;
		$cookie = new \CHttpCookie($cookieName, '');
		$cookie->httpOnly = true;
		$cookie->expire = time() + $duration;

		if (craft()->request->isSecureConnection)
		{
			$cookie->secure = true;
		}

		$cookie->value = craft()->getSecurityManager()->hashData(base64_encode(serialize($data)));
		craft()->getRequest()->getCookies()->add($cookie->name, $cookie);
	}

	/**
	 * @param $cookieName
	 * @return mixed|null
	 */
	public function getCookieValue($cookieName)
	{
		$cookie = craft()->request->getCookie($this->getStateKeyPrefix().$cookieName);

		if ($cookie && !empty($cookie->value) && ($data = craft()->securityManager->validateData($cookie->value)) !== false)
		{
			$data = @unserialize(base64_decode($data));
			return $data;
		}

		return null;
	}

	/**
	 * @return null
	 */
	public function wasSessionRestoredFromCookie()
	{
		return $this->_sessionRestoredFromCookie;
	}

	/**
	 * Clears all user identity information from persistent storage. This will remove the data stored via {@link setState}.
	 */
	public function clearStates()
	{
		if (isset($_SESSION))
		{
			$keys = array_keys($_SESSION);
			$prefix = $this->getStateKeyPrefix();

			$n = strlen($prefix);
			foreach($keys as $key)
			{
				if (!strncmp($key, $prefix, $n))
				{
					unset($_SESSION[$key]);
				}
			}
		}
	}

	/**
	 * Changes the current user with the specified identity information. This method is called by {@link login} and {@link restoreFromCookie}
	 * when the current user needs to be populated with the corresponding identity information. Derived classes may override this method
	 * by retrieving additional user-related information. Make sure the parent implementation is called first.
	 *
	 * @param mixed  $id     A unique identifier for the user
	 * @param string $name   The display name for the user
	 * @param array  $states Identity states
	 */
	protected function changeIdentity($id,$name,$states)
	{
		$this->setId($id);
		$this->setName($name);
		$this->loadIdentityStates($states);
	}

	/**
	 *
	 */
	protected function renewCookie()
	{
		$this->_checkVitals();

		$cookies = craft()->request->getCookies();
		$cookie = $cookies->itemAt($this->getStateKeyPrefix());

		// Check the identity cookie and make sure the data hasn't been tampered with.
		if ($cookie && !empty($cookie->value) && ($data = craft()->securityManager->validateData($cookie->value)) !== false)
		{
			$data = $this->getCookieValue('');

			if (is_array($data) && isset($data[0], $data[1], $data[2], $data[3], $data[4], $data[5]))
			{
				$savedUserAgent = $data[4];
				$currentUserAgent = craft()->request->userAgent;

				$this->_checkUserAgentString($currentUserAgent, $savedUserAgent);

				// Bump the expiration time.
				$cookie->expire = time() + $data[3];
				$cookies->add($cookie->name, $cookie);

				$this->authTimeout = $data[3];
			}
		}
		else
		{
			// If session duration is set to 0, then the session will be over when the browser is closed.
			if ($this->_getSessionDuration(false) > 0)
			{
				// If they are not a guest, they still have a valid PHP session, but at this point their identity cookie has expired, so let's kill it all.
				if (!$this->isGuest())
				{
					$this->logout(false);
				}
			}
		}
	}

	/**
	 * Populates the current user object with the information obtained from cookie.
	 * This method is used when automatic login ({@link allowAutoLogin}) is enabled.
	 * The user identity information is recovered from cookie.
	 */
	protected function restoreFromCookie()
	{
		$this->_checkVitals();

		// See if they have an existing identity cookie.
		$cookie = craft()->request->getCookies()->itemAt($this->getStateKeyPrefix());

		// Grab the identity cookie and make sure the data hasn't been tampered with.
		if ($cookie && !empty($cookie->value) && is_string($cookie->value) && ($data = craft()->securityManager->validateData($cookie->value)) !== false)
		{
			// Grab the data
			$data = $this->getCookieValue('');

			if (is_array($data) && isset($data[0], $data[1], $data[2], $data[3], $data[4], $data[5]))
			{
				$loginName = $data[0];
				$currentSessionToken = $data[1];
				$uid = $data[2];
				$seconds = $data[3];
				$savedUserAgent = $data[4];
				$states = $data[5];
				$currentUserAgent = craft()->request->userAgent;

				$this->_checkUserAgentString($currentUserAgent, $savedUserAgent);

				// Get the hashed token from the db based on login name and uid.
				if (($sessionRow = $this->_findSessionToken($loginName, $uid)) !== false)
				{
					$dbHashedToken = $sessionRow['token'];
					$userId = $sessionRow['userId'];

					// Make sure the given session token matches what we have in the db.
					if (craft()->security->checkString($currentSessionToken, $dbHashedToken))
					{
						// It's all good.
						if($this->beforeLogin($loginName, $states, true))
						{
							$this->changeIdentity($userId, $loginName, $states);

							if ($this->autoRenewCookie)
							{
								// Generate a new session token for the database and cookie.
								$newSessionToken = StringHelper::UUID();
								$hashedNewToken = craft()->security->hashString($newSessionToken);
								$this->_updateSessionToken($loginName, $dbHashedToken, $hashedNewToken['hash']);

								// While we're let's clean up stale sessions.
								$this->_cleanStaleSessions();

								// Save updated info back to identity cookie.
								$data = array(
									$this->getName(),
									$newSessionToken,
									$uid,
									$seconds,
									$currentUserAgent,
									$states,
								);

								$this->saveCookie('', $data, $seconds);
								$this->authTimeout = $seconds;
								$this->_sessionRestoredFromCookie = true;
								$this->_userRow = null;
							}

							$this->afterLogin(true);
						}
					}
					else
					{
						Craft::log('Tried to restore session from a cookie, but the given hashed database token value does not appear to belong to the given login name. Hashed db value: '.$dbHashedToken.' and loginName: '.$loginName.'.', LogLevel::Error);
						// Forcing logout here clears the identity cookie helping to prevent session fixation.
						$this->logout();
					}
				}
				else
				{
					Craft::log('Tried to restore session from a cookie, but the given login name does not match the given uid. UID: '.$uid.' and loginName: '.$loginName.'.', LogLevel::Error);
					// Forcing logout here clears the identity cookie helping to prevent session fixation.
					$this->logout();
				}
			}
			else
			{
				Craft::log('Tried to restore session from a cookie, but it appears we the data in the cookie is invalid.', LogLevel::Error);
				$this->logout();
			}
		}
		else
		{
			// If session duration is set to 0, then the session will be over when the browser is closed.
			if ($this->_getSessionDuration(false) > 0)
			{
				// If they are not a guest, they still have a valid PHP session, but at this point their identity cookie has expired, so let's kill it all.
				if (!$this->isGuest())
				{
					$this->logout(false);
				}
			}
		}
	}

	/**
	 * @return bool|void
	 */
	protected function beforeLogout()
	{
		$cookie = craft()->request->getCookies()->itemAt($this->getStateKeyPrefix());

		// Grab the identity cookie information and make sure the data hasn't been tampered with.
		if ($cookie && !empty($cookie->value) && is_string($cookie->value) && ($data = craft()->securityManager->validateData($cookie->value)) !== false)
		{
			// Grab the data
			$data = $this->getCookieValue('');

			if (is_array($data) && isset($data[0], $data[1], $data[2], $data[3], $data[4], $data[5]))
			{
				$loginName = $data[0];
				$uid = $data[2];

				// Clean up their row in the sessions table.
				$user = craft()->users->getUserByUsernameOrEmail($loginName);
				craft()->db->createCommand()->delete('sessions', 'userId=:userId AND uid=:uid', array('userId' => $user->id, 'uid' => $uid));
			}
			else
			{
				Craft::log('During logout, tried to remove the row from the sessions table, but it appears the cookie data is invalid.', LogLevel::Error);
			}
		}

		$this->_userRow = null;

		return true;
	}

	/**
	 * @param $loginName
	 * @param $uid
	 * @return bool
	 */
	private function _findSessionToken($loginName, $uid)
	{
		$result = craft()->db->createCommand()
		    ->select('s.token, s.userId')
		    ->from('sessions s')
		    ->join('users u', 's.userId = u.id')
		    ->where('(u.username=:username OR u.email=:email) AND s.uid=:uid', array(':username' => $loginName, ':email' => $loginName, 'uid' => $uid))
		    ->queryRow();

		if (is_array($result) && count($result) > 0)
		{
			return $result;
		}

		return false;
	}

	/**
	 * @param $loginName
	 * @param $currentToken
	 * @param $newToken
	 * @return int
	 */
	private function _updateSessionToken($loginName, $currentToken, $newToken)
	{
		$user = craft()->users->getUserByUsernameOrEmail($loginName);
		craft()->db->createCommand()->update('sessions', array('token' => $newToken), 'token=:currentToken AND userId=:userId', array('currentToken' => $currentToken, 'userId' => $user->id));
	}

	/**
	 *
	 */
	private function _cleanStaleSessions()
	{
		$interval = new DateInterval('P3M');
		$expire = DateTimeHelper::currentUTCDateTime();
		$pastTimeStamp = $expire->sub($interval)->getTimestamp();
		$pastTime = DateTimeHelper::formatTimeForDb($pastTimeStamp);

		craft()->db->createCommand()->delete('sessions', 'dateUpdated < :pastTime', array('pastTime' => $pastTime));
	}

	/**
	 * @param $rememberMe
	 * @return int
	 */
	private function _getSessionDuration($rememberMe)
	{
		if ($rememberMe)
		{
			$duration = craft()->config->get('rememberedUserSessionDuration');
		}
		else
		{
			$duration = craft()->config->get('userSessionDuration');
		}

		// Calculate how long the session should last.
		if ($duration)
		{
			$interval = new DateInterval($duration);
			$expire = DateTimeHelper::currentUTCDateTime();
			$currentTimeStamp = $expire->getTimestamp();
			$futureTimeStamp = $expire->add($interval)->getTimestamp();
			$seconds = $futureTimeStamp - $currentTimeStamp;
		}
		else
		{
			$seconds = 0;
		}

		return $seconds;
	}

	/**
	 * @param $id
	 * @return int
	 */
	private function _getUserRow($id)
	{
		if (!isset($this->_userRow))
		{
			if ($id)
			{
				$userRow = craft()->db->createCommand()
				    ->select('*')
				    ->from('{{users}}')
				    ->where('id=:id', array(':id' => $id))
				    ->queryRow();

				if ($userRow)
				{
					$this->_userRow = $userRow;
				}
				else
				{
					$this->_userRow = false;
				}
			}
			else
			{
				$this->_userRow = false;
			}
		}

		return $this->_userRow;
	}

	/**
	 * Checks whether the the current request has a user agent string or IP address.
	 */
	private function _checkVitals()
	{
		if (craft()->config->get('requireUserAgentAndIpForSession'))
		{
			// Require a userAgent string and an IP address to help prevent direct socket connections from trying to login.
			if (!craft()->request->userAgent || !craft()->request->getIpAddress())
			{
				Craft::log('Someone tried to restore a session from a cookie without presenting an IP address or userAgent string.', LogLevel::Warning);
				$this->logout(true);
				$this->requireLogin();
			}
		}
	}

	/**
	 * Checks whether the current user agent string matches the user agent string saved in the identity cookie.
	 */
	private function _checkUserAgentString($currentUserAgent, $savedUserAgent)
	{
		if (craft()->config->get('requireMatchingUserAgentForSession'))
		{
			// If the saved userAgent differs from the current one, bail.
			if ($savedUserAgent !== $currentUserAgent)
			{
				Craft::log('Tried to restore session from the the identity cookie, but the saved userAgent ('.$savedUserAgent.') does not match the current userAgent ('.$currentUserAgent.').', LogLevel::Warning);
				$this->logout(true);
			}
		}
	}
}
